use pest_derive::Parser;

#[derive(Parser)]
#[grammar = "parser/deimos.pest"]
pub struct DeimosParser;

#[cfg(test)]
mod deimos_parser_tests {
    use crate::parser::{DeimosParser, Rule};
    use pest::iterators::Pair;
    use pest::Parser;

    macro_rules! fail_rule {
        ($pair: expr; $rule:path) => {{
            panic!(
                "Expected {} but found {:?}",
                stringify!($rule),
                $pair.as_rule()
            )
        }};
    }

    macro_rules! match_rule {
        ($pair:expr; $rule:path) => {{
            if $rule != $pair.as_rule() {
                fail_rule!($pair; $rule);
            }
        }};
    }

    macro_rules! match_inner_rules {
        ($pair:expr; $rule:path; $handle_last:expr) => {{
            if let $rule = $pair.as_rule() {
                $handle_last($pair);
            } else {
                fail_rule!($pair; $rule);
            }
        }};
        ($pair:expr; $head_rule:path $(, $tail_rules:path )* ; $handle_last:expr) => {{
            if let $head_rule = $pair.as_rule() {
                let inner = $pair.into_inner().next().unwrap();
                match_inner_rules!(
                    inner;
                    $( $tail_rules ),*;
                    $handle_last
                );
            } else {
                fail_rule!($pair; $head_rule);
            }
        }};
    }

    fn parse(rule: Rule, input: &str) -> Pair<Rule> {
        let pair = DeimosParser::parse(rule, input)
            .expect("Parsing failed.")
            .next()
            .unwrap();
        dbg!(&pair);
        pair
    }

    #[test]
    fn hex_int() {
        let pair = parse(Rule::IntLiteral, "0x1234abcd");
        match_inner_rules!(pair; Rule::IntLiteral, Rule::NumberBase, Rule::HexadecimalBase; |hexadecimal_base: Pair<Rule>| {
            assert_eq!(hexadecimal_base.as_str(), "0x1234abcd")
        });
    }

    #[test]
    fn hex_long() {
        let pair = parse(Rule::LongLiteral, "0x123456789abcdefL");
        match_inner_rules!(pair; Rule::LongLiteral, Rule::NumberBase, Rule::HexadecimalBase; |hexadecimal_base: Pair<Rule>| {
            assert_eq!(hexadecimal_base.as_str(), "0x123456789abcdef")
        })
    }

    #[test]
    fn identifier_call_as_expression() {
        let pair = parse(Rule::Expression, "foo()");
        match_inner_rules!(pair; Rule::Expression, Rule::OrExpression, Rule::AndExpression, Rule::EqualityExpression, Rule::ComparisonExpression, Rule::AdditiveExpression, Rule::MultiplicativeExpression, Rule::UnaryExpression; |unary_expression: Pair<Rule>| {
            let mut unary_pairs = unary_expression.into_inner();
            let primary_expression = unary_pairs.next().unwrap();
            match_rule!(primary_expression; Rule::PrimaryExpression);
            let call = unary_pairs.next().unwrap();
            match_rule!(call; Rule::Call);
        });
    }

    #[test]
    fn identifier_call_as_call_expression() {
        let pair = parse(Rule::CallExpression, "foo()");
        match_rule!(pair; Rule::CallExpression);
        let mut call_expression_pairs = pair.into_inner();
        let primary_expression = call_expression_pairs.next().unwrap();
        match_rule!(primary_expression; Rule::PrimaryExpression);
        let call = call_expression_pairs.next().unwrap();
        match_rule!(call; Rule::Call);
    }

    mod smoke_screen_tests {
        use crate::parser::deimos_parser_tests::parse;
        use crate::parser::Rule;

        #[test]
        fn simple_interface() {
            parse(Rule::CompilationUnit, "pub int Simple { fn foo() -> Void }");
        }

        #[test]
        fn interface_with_op() {
            parse(
                Rule::CompilationUnit,
                "pub int Callable { op () () -> Void }",
            );
        }

        #[test]
        fn interface_with_alias() {
            parse(
                Rule::CompilationUnit,
                "pub int Callable {\n  fn call() -> Void\n  op () () -> Void alias call\n}",
            );
        }
    }
}
